@@ -1,6 +1,6 @@
 import torch
 from functools import partial
-from torch import nn
+from torch import nn, einsum
 import torch.nn.functional as F
 
 try:
@@ -27,6 +27,15 @@ def eval_decorator(fn):
         return out
     return inner
 
+def non_cosine_sim_attn_fn(q, k, v, **kwargs):
+    q = q * (q.shape[-1] ** -0.5)
+    sim = einsum('b h i d, b h j d -> b h i j', q, k)
+    i, j = sim.shape[-2:]
+    causal_mask = torch.ones((i, j), dtype = torch.bool, device = q.device).triu(j - i + 1)
+    sim = sim.masked_fill(causal_mask, -torch.finfo(sim.dtype).max)
+    attn = sim.softmax(dim = -1)
+    return einsum('b h i j, b h j d -> b h i d', attn, v)
+
 # top k filtering
 
 def top_k(logits, thres = 0.9):
@@ -38,9 +47,10 @@ def top_k(logits, thres = 0.9):
 
 # attention and feedforward
 
-def FeedForward(dim, mult = 4):
+def FeedForward(dim, mult = 4, pre_norm = False):
     dim_hidden = int(dim * mult)
     return nn.Sequential(
+        nn.LayerNorm(dim) if pre_norm else nn.Identity(),
         nn.Linear(dim, dim_hidden, bias = False),
         nn.GELU(),
         nn.Linear(dim_hidden, dim, bias = False)
@@ -54,16 +64,27 @@ class Attention(nn.Module):
         heads = 8,
         scale = 8,
         l2norm_groups = 1,
+        pre_norm = False,
         use_cuda_kernel = False,
+        non_cosine_sim_attn = False,
         **kwargs
     ):
         super().__init__()
         inner_dim = dim_head * heads
+
+        self.norm = nn.LayerNorm(dim) if pre_norm else nn.Identity()
+
         self.scale = scale
         self.heads = heads
 
         self.l2norm_groups = l2norm_groups
-        self.attn_fn = plain_cosine_sim_attention if not use_cuda_kernel else partial(flash_cosine_sim_attention, **kwargs)
+
+        if non_cosine_sim_attn:
+            self.attn_fn = non_cosine_sim_attn_fn
+        elif use_cuda_kernel:
+            self.attn_fn = partial(flash_cosine_sim_attention, **kwargs)
+        else:
+            self.attn_fn = plain_cosine_sim_attention
 
         self.to_q = nn.Linear(dim, inner_dim, bias = False)
         self.to_k = nn.Linear(dim, inner_dim, bias = False)
@@ -73,6 +94,8 @@ class Attention(nn.Module):
     def forward(self, x):
         h, scale, l2norm_groups = self.heads, self.scale, self.l2norm_groups
 
+        x = self.norm(x)
+
         q, k, v = self.to_q(x), self.to_k(x), self.to_v(x)
         q, k, v = map(lambda t: rearrange(t, 'b n (h d) -> b h n d', h = h), (q, k, v))
 
@@ -96,6 +119,8 @@ class CosineSimCausalTransformer(nn.Module):
         heads = 8,
         dim_head = 64,
         use_cuda_kernel = False,
+        pre_norm = False,
+        non_cosine_sim_attn = False,
         **kwargs
     ):
         super().__init__()
@@ -104,20 +129,24 @@ class CosineSimCausalTransformer(nn.Module):
         self.token_emb = nn.Embedding(num_tokens, dim)
         self.pos_emb = nn.Embedding(max_seq_len, dim)
 
-        self.residual_scale = (2 * depth) ** 0.25
+        self.residual_scale = 1 if pre_norm else ((2 * depth) ** 0.25)
 
         self.layers = nn.ModuleList([])
         for _ in range(depth):
             self.layers.append(nn.ModuleList([
-                Attention(dim, dim_head = dim_head, heads = heads, use_cuda_kernel= use_cuda_kernel, scale = attn_scale, groups = attn_l2norm_groups, **kwargs),
-                nn.LayerNorm(dim),
-                FeedForward(dim),
-                nn.LayerNorm(dim),
+                Attention(dim, dim_head = dim_head, heads = heads, use_cuda_kernel= use_cuda_kernel, scale = attn_scale, groups = attn_l2norm_groups, pre_norm = pre_norm, non_cosine_sim_attn = non_cosine_sim_attn, **kwargs),
+                nn.LayerNorm(dim) if not pre_norm else nn.Identity(),
+                FeedForward(dim, pre_norm = pre_norm),
+                nn.LayerNorm(dim) if not pre_norm else nn.Identity(),
             ]))
 
-        self.to_logits = nn.Linear(dim, num_tokens, bias = False)
+        self.to_logits = nn.Sequential(
+            nn.LayerNorm(dim) if pre_norm else nn.Identity(),
+            nn.Linear(dim, num_tokens, bias = False)
+        )
 
-        self.init_(depth)
+        if not pre_norm:
+            self.init_(depth)
 
     def init_(self, depth):
         nn.init.normal_(self.token_emb.weight, std = 1e-5)
@@ -130,10 +159,10 @@ class CosineSimCausalTransformer(nn.Module):
             init_weight_xavier_normal_(attn.to_k, 1.)
             init_weight_xavier_normal_(attn.to_v, init_gain)
             init_weight_xavier_normal_(attn.to_out, init_gain)
-            init_weight_xavier_normal_(ff[0], init_gain)
-            init_weight_xavier_normal_(ff[2], init_gain)
+            init_weight_xavier_normal_(ff[1], init_gain)
+            init_weight_xavier_normal_(ff[3], init_gain)
 
-        init_weight_xavier_normal_(self.to_logits, 1)
+        init_weight_xavier_normal_(self.to_logits[-1], 1)
 
     @torch.no_grad()
     @eval_decorator
